import torch

from typing import List, Optional, Tuple
from pytorch3d.renderer.cameras import CamerasBase


def sample_pdf(depth_values, weights, N_samples_fine, det=False):
    """ 
    # reference: https://github.com/bmild/nerf/blob/18b8aebda6700ed659cb27a0c348b737a5f6ab60/run_nerf_helpers.py#L183
    
    depth_values: n_rays, n_pts
    weights: n_rays, n_pts, 1
    
    """
    
    device = depth_values.device
    bins = .5 * (depth_values[:, 1:] + depth_values[:, :-1])
    
    weights = weights.squeeze(-1) # n_rays, n_pts
    weights += 1e-5
    pdf = weights / torch.sum(weights, dim=1, keepdim=True)
    cdf = torch.cumsum(pdf, dim=1)
    cdf = torch.cat([torch.zeros_like(cdf[:, :1]), cdf], dim=1)
    
    # Take uniform samples
    if det:
        u = torch.linspace(0., 1., N_samples_fine)
        # import pdb; pdb.set_trace()
        u = torch.broadcast_to(u, list(cdf.shape[:-1]) + [N_samples_fine])
    else:
        u_size = list(cdf.shape[:1]) + [N_samples_fine]
        u = torch.rand(u_size, device=device) # n_rays, n_pts

    # Invert CDF
    inds = torch.searchsorted(cdf, u, right=True) # first N_dims should match between cdf and u
    below = torch.maximum(torch.zeros_like(inds), inds-1)
    above = torch.minimum(torch.ones_like(inds) * (cdf.shape[1]-1), inds)
    # inds_g = torch.stack([below, above], -1)
    # cdf_g = torch.gather(cdf, inds_g, axis=-1, batch_dims=len(inds_g.shape)-2)
    # bins_g = torch.gather(bins, inds_g, axis=-1, batch_dims=len(inds_g.shape)-2)
    
    # https://stackoverflow.com/questions/50999977/what-does-the-gather-function-do-in-pytorch-in-layman-terms
    cdf_g_below = torch.gather(cdf, 1, below)
    cdf_g_above = torch.gather(cdf, 1, above)
    bins_g_below = torch.gather(bins, 1, below)
    bins_g_above = torch.gather(bins, 1, above)

    # denom = (cdf_g[..., 1]-cdf_g[..., 0])
    # denom = torch.where(denom < 1e-5, torch.ones_like(denom), denom)
    # t = (u-cdf_g[..., 0])/denom
    # samples = bins_g[..., 0] + t * (bins_g[..., 1]-bins_g[..., 0])
    denom = cdf_g_above - cdf_g_below
    denom = torch.where(denom < 1e-5, torch.ones_like(denom), denom)
    t = (u-cdf_g_below)/denom
    samples = bins_g_below + t * (bins_g_above-bins_g_below) # z_val_samples

    return samples
    

# Volume renderer which integrates color and density along rays
# according to the equations defined in [Mildenhall et al. 2020]
class VolumeRenderer(torch.nn.Module):
    def __init__(
        self,
        cfg,
        cfg_implicit=None,
    ):
        super().__init__()

        self._chunk_size = cfg.chunk_size
        self._white_background = cfg.white_background if 'white_background' in cfg else False
        
        self.use_fine = cfg.use_fine
        if self.use_fine:
            from implicit import NeuralRadianceField
            self.implicit_fine = NeuralRadianceField(cfg_implicit)
            self.N_samples_fine = 128 # coarse: 64, fine: 128
        
    def _compute_weights(
        self,
        deltas,
        rays_density: torch.Tensor,
        eps: float = 1e-10
    ):
        # TODO (1.5): Compute transmittance using the equation described in the README
        n_rays, n_pts, _ = rays_density.shape
        device = rays_density.device
        transmittance = torch.exp(-torch.cumsum(rays_density * deltas, dim=1))
        transmittance = torch.cat([torch.ones(n_rays, 1, 1, device=device), transmittance[:, :-1]], dim=1)
        # import pdb; pdb.set_trace()
        # import pdb; pdb.set_trace()

        # TODO (1.5): Compute weight used for rendering from transmittance and density
        weights = transmittance * (1 - torch.exp(-rays_density * deltas))
        # import pdb; pdb.set_trace()
        return weights
    
    def _aggregate(
        self,
        weights: torch.Tensor,
        rays_feature: torch.Tensor
    ):
        # TODO (1.5): Aggregate (weighted sum of) features using weights
        n_rays, n_pts, _ = weights.shape
        rays_feature = rays_feature.view(n_rays, n_pts, 3)

        feature = (rays_feature * weights).sum(1)
        # import pdb; pdb.set_trace()

        return feature

    def forward(
        self,
        sampler,
        implicit_fn,
        ray_bundle,
    ):
        B = ray_bundle.shape[0]

        # Process the chunks of rays.
        chunk_outputs = []
        
        # from render_functions import render_points
        # n_pts = .sample_shape[1]
        # ray_bundle = sampler(ray_bundle)
        # points = ray_bundle.sample_points
        # render_points('results/pts1.png', points.reshape(-1, 3).unsqueeze(0))

        for chunk_start in range(0, B, self._chunk_size):
        # for chunk_start in range(0, self._chunk_size, B):
            cur_ray_bundle = ray_bundle[chunk_start:chunk_start+self._chunk_size]
            # cur_ray_bundle = ray_bundle[chunk_start:chunk_start+B]

            # Sample points along the ray
            cur_ray_bundle = sampler(cur_ray_bundle)
            
            n_pts = cur_ray_bundle.sample_shape[1]
            # points = cur_ray_bundle.sample_points
            # # points = (points - points.min()) / (points.max() - points.min()) - .5
            # # import pdb; pdb.set_trace()
            # render_points('results/pts.png', points.reshape(-1, 3).unsqueeze(0))

            # Call implicit function with sample points
            implicit_output = implicit_fn(cur_ray_bundle)
            density = implicit_output['density']
            feature = implicit_output['feature']

            # Compute length of each ray segment
            depth_values = cur_ray_bundle.sample_lengths[..., 0]
            deltas = torch.cat(
                (
                    depth_values[..., 1:] - depth_values[..., :-1],
                    1e10 * torch.ones_like(depth_values[..., :1]),
                ),
                dim=-1,
            )[..., None]

            # Compute aggregation weights
            weights = self._compute_weights(
                deltas.view(-1, n_pts, 1),
                density.view(-1, n_pts, 1)
            )
            
            if self.use_fine:
                z_val_samples = sample_pdf(depth_values, weights[:, 1:-1], self.N_samples_fine)
                z_val_samples = z_val_samples.detach()
                
                z_vals, _ = torch.sort(torch.cat([depth_values, z_val_samples], dim=-1), dim=-1)
                # update sample points and sample lengths
                rays_o = cur_ray_bundle.origins
                rays_d = cur_ray_bundle.directions
                sample_points = rays_o.unsqueeze(1) + rays_d.unsqueeze(1) * z_vals.unsqueeze(-1)
                cur_ray_bundle = cur_ray_bundle._replace(
                    sample_points=sample_points,
                    sample_lengths=z_vals.unsqueeze(-1),
                )
                n_pts = cur_ray_bundle.sample_shape[1]
                
                # Call implicit function with sample points
                # implicit_output = implicit_fn(cur_ray_bundle)
                implicit_output = self.implicit_fine(cur_ray_bundle)
                density = implicit_output['density']
                feature = implicit_output['feature']
                
                # Compute length of each ray segment
                depth_values = cur_ray_bundle.sample_lengths[..., 0]
                deltas = torch.cat(
                    (
                        depth_values[..., 1:] - depth_values[..., :-1],
                        1e10 * torch.ones_like(depth_values[..., :1]),
                    ),
                    dim=-1,
                )[..., None]

                # Compute aggregation weights
                weights = self._compute_weights(
                    deltas.view(-1, n_pts, 1),
                    density.view(-1, n_pts, 1)
                )
                
            # TODO (1.5): Render (color) features using weights
            feature = self._aggregate(weights, feature)

            # TODO (1.5): Render depth map
            # import pdb; pdb.set_trace()
            depth = depth_values.unsqueeze(-1) * weights
            depth = depth.sum(1)

            # Return
            cur_out = {
                'feature': feature,
                'depth': depth,
            }

            chunk_outputs.append(cur_out)

        # Concatenate chunk outputs
        out = {
            k: torch.cat(
              [chunk_out[k] for chunk_out in chunk_outputs],
              dim=0
            ) for k in chunk_outputs[0].keys()
        }

        return out


renderer_dict = {
    'volume': VolumeRenderer,
    # 'volume-fine': VolumeRenderer
}
